%% PICEA-g MATLAB code
%**************************************************************************
% Main Idea:
% PICEA-g is developed to solve multi/many-objective problems. It co-evolves
% candidate solutions with a set of goal vectors during the seach progress. 
% Candiate solutions are guided twoards the Pareto optimal front by goal vectors.
% Goal vectors gain fitness by offering comparablity to the candidate solutions

% The PICEA-g employs the "mu+lamada" elitist framework, that is, select
% the best N solutions from a union of parent (N) and offspring (N) 
% population
%**************************************************************************
% Reference: 
% R. Wang, R. Purshouse, and P. Fleming, Preference-inspired
% coevolutionary algorithms for many-objective optimisation, IEEE Trans.
% Evol. Comput., 2013. In press.
%**************************************************************************
%
% <1> Note that in this version, goal vector bounds are estimated as follows  
%     goalUpper = alpha*curnadir;
%     goalLower = curideal;
%     where curnadir and curideal are estimated ideal and nadir point

% <2> As the effect of Pareto dominance check is not significant
%     for many-objective problems, the Pareto dominance check specifed 
%     in the paper is not applied here. Also the fitness assignment used in
%     PICEA-g is weak Pareto dominace complied.

% <3> Note that this code demonstrates the performance of PICEA-g on 
%     2-objective WFG4 problem
%
% <4> For any given problem, the algorithm only requires population size (N)
%     maximum generations (maxGen) as input parameters
%**************************************************************************

% Author:     Rui Wang     The University of Sheffield, UK & 
%                          National University of Defense Technology, P.R.China 
% History:    02.OCT.2011  file created
%             04.SEP.2015  clean up
%             19.June.2016 add normalization procedure 
  
% Feel free to contact me if you have any questions on this algorithm
% Dr. Rui Wang,   Email: ruiwangnudt@gmail.com
% Copyright reserved by the authors

%% ************************************************
clear all; close all;
% The number of objectives
objvNo = 2;
% testNo: test problem
testNo = 4;  
% The population size of candidate solutions and goal vectors
N = 100; Ngoal = N; 
% The maximum number of generations
maxGen = 250;

%% create figure
hf = figure;
%% parameters for decision variables. 
% ******** Note that this might need to changed to follow your own problem ****** %
% i) WFG test problem setting
k = 4;  l = 4; % the wfg problem parameters
nVar = k + l;  % number of decision variables
bounds=[zeros(1, nVar); 2*(1:nVar)];  % bounds of decision varialbes

% ii) DTLZ test problem setting
% nVar = 20;  % number of decision variables
% k = nVar-objvNo+1; % DTLZ problem parameter
% bounds=[zeros(1, nVar); ones(1,nVar)];  % bounds of decision varialbes

%% initialise offline archive
% To store the best so far Pareto optimal front
bestobjv = Inf*ones(1,objvNo);
% To store the best so far Pareto optimal set 
bestphen = NaN*ones(1,nVar);

%% initialize candidate solutions
P = crtrp(N,bounds);

%% compute objective values of P
% ******** Note that you can replace "wfg" with your own objective functions *******%
F_P = wfgextend(P, objvNo, k, l, testNo); 
% F_P = dtlz(P, objvNo,k,testNo);

%% obtain estimated ideal and nadir points
Cobjvix = find_nd(F_P);
Hobjv =  F_P(logical(Cobjvix),:);
curideal = min(Hobjv,[],1);
curnadir = max(Hobjv,[],1);

%% generate goal vectors within goal vector bounds
alpha = 1.2; % suggested to be set within (1,2)
goalUpper = alpha*curnadir;  goalLower = curideal;
Goal = initGoals(Ngoal,goalUpper,goalLower); 

%% Evolution process
for gen = 1:maxGen
   %% obtain new solutions by crossover and mutation
    % shuffle candidate solutions first
    numn = size(P,1);  rx = randperm(numn); C = P(rx,:);
    % apply crrossover (SBX) and mutation (PM) operators to generate new offspring 
    % Note that SBX and PM is only for real-number optimization.  You might need to 
    % design your own crossover and mutation operators
    C= sbx_sal(C, bounds, 15, 0, 1, 1,0.5);
    C = polymut_sal(C, bounds, 20, 1/nVar);
    
    %% Compute objective values for the new offspring 
    % For your problem you should also replace "wfg" with your own objective functions 
    F_C = wfgextend(C, objvNo, k, l, testNo);
    %F_C = dtlz(C, objvNo,k,testNo);
    
    %% combine the parent and offspring population for candidate solutions
    jointP = [P;C]; jointF = [F_P;F_C];
    
    %% obtain non-dominated solutions
    Cobjvix = find_nd(jointF);
    Hobjv =  jointF(logical(Cobjvix),:);
    HP = jointP(logical(Cobjvix),:);
    [ix,bestix] = find_nd(Hobjv,bestobjv);
    bestobjv = [bestobjv(logical(bestix),:) ; Hobjv(logical(ix),:)];
    bestphen = [bestphen(logical(bestix),:) ; HP(logical(ix),:)];
    
   %% update current ideal and nadir point
    [curideal,curnadir] = obtainIdealNadir(bestobjv,curideal,curnadir);
   
   %% prune bestobjv to a fixed size using SPEA clustering technique (IF NECESSARY)
    bestNum = size(bestobjv,1);
    if size(bestobjv,1)>N
        [bestobjv,index] = reducer(bestobjv, objvNo, N);
        bestphen = bestphen(index,:);
    end
    
    %% generate new goal vectors
    goalUpper = alpha*curnadir;  goalLower = curideal;
    GoalC = initGoals(Ngoal,goalUpper,goalLower); 
   
    %%  combine the parent and offspring population for goal vectors
    jointG = [Goal; GoalC];
    
    %% fitness assignment <the core part of the PICEA-g>
    result = fitness_PICEAg(jointF,jointP,jointG,objvNo);
    scoreS = result{1};  scoreG = result{2};
    
    %% select the best N candidate solutions and Ngoal goal vectors
    ixS = trunc_No(scoreS,N);
    P = jointP(ixS,:);
    F_P = jointF(ixS,:);
    ix = trunc_No(scoreG,Ngoal);
    Goal = jointG(ix,:);
    
    %% Plot the results
    CurrentPF=bestobjv;
    CurrentPS=bestphen;
    %plot(CurrentPF(:,1),CurrentPF(:,2),'b*',Goal(:,1),Goal(:,2),'ro'); legend('Pareto optimal front','Goal vectors')
    plot(CurrentPF(:,1),CurrentPF(:,2),'b*');
    title({['TestNo: ',num2str(testNo)],['Generation: ',num2str(gen)]})
    drawnow
end


    
   
    
  
